﻿using System.Collections;
using System.Collections.Generic;

using System.IO;
using System.Linq;

using UnityEditor;

using UnityEngine;
using UnityEngine.Networking;

public class FolderToAssetBundle
{

    private const string LocalAssetBundleOutputPath = "Assets/+AssetBundles";

    private static readonly string ProjectPath = Application.dataPath;
    
    private static readonly char Slash = Path.DirectorySeparatorChar;

    /// <summary>
    ///     Flips any instances of the wrong slash in a path (i.e. \ instead of /) to the correct slash for this platform.
    /// </summary>
    private static string CorrectDirectorySeperators
        (string directory) =>
            directory.Replace(oldChar: Path.AltDirectorySeparatorChar,
                              newChar: Path.DirectorySeparatorChar);
    
    /// <summary>
    ///     Extracts the end folder name from a path
    /// </summary>
    private static string ExtractLastPartOfPath
        (string directory)
    {
        var lastSlashIndex = directory.LastIndexOf(value: Path.DirectorySeparatorChar) + 1;

        return directory.Substring(startIndex: lastSlashIndex,
                                   length: directory.Length - lastSlashIndex);
    }

    /// <summary>
    ///     Turns a full path (i.e. D:/Projects/etc) into a project-relative path (i.e. one that starts from the Assets folder)
    /// </summary>
    private static string LocalizeFullPath
        (string directory) =>
            $"Assets{Slash}{directory.Remove(startIndex: 0, count: ProjectPath.Length + 1)}";

    /// <summary>
    ///     Turns a project-relative path (i.e. one that starts from the Assets folder) into a full path (i.e. D:/Projects/etc)
    /// </summary>
    private static string GlobalizeLocalPath
        (string directory) =>
            $"{ProjectPath.Substring(startIndex: 0, length: ProjectPath.Length - 6)}{directory}";

    // ------------------------------------------------------------------------------------------------ Context Menu //

    [MenuItem(itemName: "Assets/Export Folder As AssetBundle",
        isValidateFunction: true)]
    private static bool NewMenuOptionValidation()
    {
        // This Context Menu option will be greyed out if this method returns false
        return Directory.Exists(path: AssetDatabase.GetAssetPath(assetObject: Selection.activeObject));
    }


    [MenuItem(itemName: "Assets/Export Folder As AssetBundle")]
    private static void ExportFolderAsAssetBundle()
    {
        var path = GlobalizeLocalPath(directory: AssetDatabase.GetAssetPath(assetObject: Selection.activeObject));

        path = CorrectDirectorySeperators(directory: path);

        Debug.Log(message: $"Full Path [{path}]");

        Debug.Log(message: $"Local Path [{LocalizeFullPath(directory: path)}]");

        Debug.Log(message: $"Bundle Name {ExtractLastPartOfPath(directory: path)}");

        RecursiveAssetBundlerDrill(startingDirectory: path,
                                   bundleName: ExtractLastPartOfPath(directory: path));

        AssetDatabase.RemoveUnusedAssetBundleNames();

        if (!Directory.Exists(path: LocalAssetBundleOutputPath))
        {
            Directory.CreateDirectory(path: LocalAssetBundleOutputPath);
        }

        var outputPath = GlobalizeLocalPath(directory: $"Assets{Slash}+AssetBundles");

        outputPath = CorrectDirectorySeperators(directory: outputPath);

//        BuildPipeline.BuildAssetBundles(outputPath: outputPath,
//                                        assetBundleOptions: BuildAssetBundleOptions.ChunkBasedCompression,
//                                        targetPlatform: EditorUserBuildSettings.activeBuildTarget);

        var bundleName = ExtractLastPartOfPath(directory: path);

//        BuildPipeline.P
        
        BuildPipeline.BuildAssetBundles(outputPath: outputPath,
                                        builds: new[]
                                                {
                                                    new AssetBundleBuild
                                                    {
                                                        assetBundleName = bundleName,
                                                        assetNames =
                                                            AssetDatabase.GetAssetPathsFromAssetBundle(
                                                                assetBundleName: bundleName)
                                                    }
                                                },
                                        assetBundleOptions: BuildAssetBundleOptions.ChunkBasedCompression,
                                        targetPlatform: EditorUserBuildSettings.activeBuildTarget);
        
        AssetDatabase.Refresh();
    }

    // ----------------------------------------------------------------------------------------------------- Helpers //
    
    /// <summary>
    ///     This function will iterate through all files in all subdirectories of the given path and,
    ///     if the file isn't a meta file (which aren't required in AssetBundles?), will assign it to a
    ///     bundle of the given bundleName.
    /// </summary>
    ///
    /// <param name="startingDirectory">
    ///     What directory should we iterate through?
    /// </param>
    ///
    /// <param name="bundleName">
    ///    What name should this new bundle go by?
    /// </param>
    private static void RecursiveAssetBundlerDrill
    (
        string startingDirectory,
        string bundleName)
    {
        foreach (var directory in Directory.GetDirectories(path: startingDirectory))
        {
            foreach (var filename in Directory.GetFiles(path: directory))
            {
                if (filename.EndsWith(value: ".meta")) continue;
				
                Debug.Log(message: $"Working on asset at path [{filename}]");

                var localizedAssetPath = LocalizeFullPath(directory: filename);
				
                var assetImporter = AssetImporter.GetAtPath(path: localizedAssetPath);

                if (assetImporter == null)
                {
                    Debug.LogWarning(message: $"Getting the AssetImporter for asset at path [{localizedAssetPath}] failed.");

                    continue;
                }

                assetImporter.assetBundleName = bundleName;

                assetImporter.SaveAndReimport();
            }

            RecursiveAssetBundlerDrill(startingDirectory: directory,
                                      bundleName: bundleName);
        }
    }

    public static List<string> GetAllAssetsFromDirectory
    (
        string       directory)
    {
        var output = new List<string>();

        RecursiveAssetGetter(startingDirectory: directory,
                             assets: output);

        return output;
    }


    private static void RecursiveAssetGetter
    (
        string startingDirectory,
        List<string> assets)
    {
        foreach (var directory in Directory.GetDirectories(path: startingDirectory))
        {
            foreach (var filename in Directory.GetFiles(path: directory))
            {
                if (filename.EndsWith(value: ".meta")) continue;
				
                Debug.Log(message: $"Collecting asset [{filename}]");

                assets.Add(item: ExtractLastPartOfPath(directory: filename));
            }

            RecursiveAssetGetter(startingDirectory: directory, 
                                 assets: assets);
        }
    }
}