using UnityEngine;

namespace Vanilla.Core.Extensions
{

	public static class FloatExtensions
	{

		// ------------------------------------------------------------------------------------------ Clamping - Ref //

		/// <summary>
		/// 	Returns a float clamped between the given minimum and maximum.
		/// </summary>
		public static void Clamp(ref this float input,
		                         float          min,
		                         float          max) => input = Mathf.Clamp(value: input,
		                                                                    min: min,
		                                                                    max: max);


		public static void Clamp(ref this float input,
		                         FloatRange     range) => input = Mathf.Clamp(value: input,
		                                                                      min: range.min,
		                                                                      max: range.max);


		public static void Clamp01(ref this float input) => input = Mathf.Clamp01(value: input);


		// -------------------------------------------------------------------------------------- Clamping - Returns //

		
		public static float GetClamp(this float input,
		                             float      min,
		                             float      max) => Mathf.Clamp(value: input,
		                                                            min: min,
		                                                            max: max);


		public static float GetClamp(this float input,
		                             FloatRange range) => Mathf.Clamp(value: input,
		                                                              min: range.min,
		                                                              max: range.max);


		public static float GetClamp01(this float input) => Mathf.Clamp01(value: input);


		// ------------------------------------------------------------------------------------------ Wrapping - Ref //

		public static void Wrap(this ref float input,
		                            float      max) => Mathf.Clamp(value: input - Mathf.Floor(f: input / max) * max,
		                                                           min: 0.0f,
		                                                           max: max);


		public static void Wrap(this ref float input,
		                            float      min,
		                            float      max) => Mathf.Clamp(value: input - Mathf.Floor(f: input / max) * max,
		                                                           min: min,
		                                                           max: max);


		public static void Wrap(this ref float input,
		                            FloatRange range) => Mathf.Clamp(value: input - Mathf.Floor(f: input / range.max) * range.max,
		                                                             min: range.min,
		                                                             max: range.max);
		
		// -------------------------------------------------------------------------------------- Wrapping - Returns //


		public static float GetWrap(this float input,
		                            float      max) => Mathf.Clamp(value: input - Mathf.Floor(f: input / max) * max,
		                                                           min: 0.0f,
		                                                           max: max);


		public static float GetWrap(this float input,
		                            float      min,
		                            float      max) => Mathf.Clamp(value: input - Mathf.Floor(f: input / max) * max,
		                                                           min: min,
		                                                           max: max);


		public static float GetWrap(this float input,
		                            FloatRange range) => Mathf.Clamp(value: input - Mathf.Floor(f: input / range.max) * range.max,
		                                                             min: range.min,
		                                                             max: range.max);
		

		// ---------------------------------------------------------------------------------------------- Validation //
		
		// Sign
		
		public static bool IsPositive(this float input) => input > 0.0f;

		public static bool IsPositiveOrZero(this float input) => input >= 0.0f;
		
		public static bool IsNegative(this float input) => input < 0.0f;

		public static bool IsNegativeOrZero(this float input) => input <= 0.0f;
		
		// Normalization

		public static bool IsNormalized(this float input) => !( input < 0 || input > 1.0f );

		// Inclusive

		public static bool IsWithinInclusiveRange(this float input,
		                                          float      max) => input >= 0.0f && input <= max;

		public static bool IsWithinInclusiveRange(this float input,
		                                          float      min,
		                                          float      max) => input >= min && input <= max;

		public static bool IsWithinInclusiveRange(this float input,
		                                          FloatRange range) => input >= range.min && input <= range.max;

		// Exclusive
		
		public static bool IsWithinExclusiveRange(this float input,
		                                          float      max) => input > 0.0f && input < max;


		public static bool IsWithinExclusiveRange(this float input,
		                                          float      min,
		                                          float      max) => input > min && input < max;

		public static bool IsWithinExclusiveRange(this float input,
		                                          FloatRange range) => input > range.min && input < range.max;
		

		// --------------------------------------------------------------------------------- Normalization - Returns //

		/// <summary>
		/// 	Returns a float representing a normalized value (0-1) between the input and the given maximum.
		/// </summary>
		public static float GetNormal(this float input,
		                              float      max) => Mathf.Clamp01(value: input / max);


		// ------------------------------------------------------------------------------------------ Rounding - Ref //

		/// <summary>
		/// 	Sets the float to the nearest whole number.
		/// </summary>
		public static void ToRound(ref this float input) => input = Mathf.Round(f: input);


		/// <summary>
		/// 	Sets the float to the next highest whole number.
		/// </summary>
		public static void ToCeiling(ref this float input) => input = Mathf.Ceil(f: input);


		/// <summary>
		/// 	Sets the float to the next lowest whole number.
		/// </summary>
		public static void ToFloor(ref this float input) => input = Mathf.Floor(f: input);


		/// <summary>
		/// 	Sets the float to the nearest multiple of factor.
		///
		/// 	This function will be able to handle positive or negative values.
		/// </summary>
		public static void ToNearest(ref this float input,
		                             float          factor) => 
			input = Mathf.Round(f: Mathf.Abs(f: input) / factor) * ( input.IsPositive() ? factor : -factor );


		/// <summary>
		/// 	Sets the float to the nearest multiple of factor.
		///
		/// 	This function is only be able to handle positive values.
		/// </summary>
		public static void ToNearestPositive(ref this float input,
		                                     float          factor) => input = Mathf.Round(f: input / factor) * factor;


		// -------------------------------------------------------------------------------------- Rounding - Returns //

		
		/// <summary>
		/// 	Returns a float of the nearest whole number.
		/// </summary>
		public static float GetRound(this float input) => Mathf.Round(f: input);


		/// <summary>
		/// 	Returns a float to the next highest whole number.
		/// </summary>
		public static float GetCeiling(this float input) => Mathf.Ceil(f: input);


		/// <summary>
		/// 	Returns a float to the next lowest whole number.
		/// </summary>
		public static float GetFloor(this float input) => Mathf.Floor(f: input);


		// Rounding to integers
		/// <summary>
		/// 	Returns an integer the nearest multiple of factor.
		///
		/// 	This function will be able to handle positive or negative values.
		/// </summary>
		public static int GetNearestInt(ref this float input,
		                                float          factor) => Mathf.RoundToInt(f: Mathf.Round(f: Mathf.Abs(f: input) / factor) * ( input.IsPositive() ? factor : -factor ));

		/// <summary>
		/// 	Returns an integer the nearest multiple of factor.
		///
		/// 	This function is only be able to handle positive values.
		/// </summary>
		public static int GetNearestPositiveInt(ref this float input,
		                                        float          factor) => Mathf.RoundToInt(f: Mathf.Round(f: input / factor) * factor);


		/// <summary>
		/// 	Returns an integer of the nearest whole number.
		/// </summary>
		public static int GetRoundInt(this float input) => Mathf.RoundToInt(f: input);


		/// <summary>
		/// 	Returns an integer of the next highest whole number.
		/// </summary>
		public static int GetCeilingInt(ref this float input) => Mathf.CeilToInt(f: input);


		/// <summary>
		/// 	Returns an integer of the next lowest whole number.
		/// </summary>
		public static int GetFloorInt(ref this float input) => Mathf.FloorToInt(f: input);


		/// <summary>
		/// 	Returns a float to the nearest value.
		///
		/// 	This function will be able to handle positive or negative values.
		/// </summary>
		public static float GetToNearest(this float input,
		                                 float      factor) => Mathf.Round(f: Mathf.Abs(f: input) / factor) * ( input.IsPositive() ? factor : -factor );


		/// <summary>
		/// 	Returns a float to the nearest value.
		///
		/// 	This function is only able to handle positive values.
		/// </summary>
		public static float GetNearestPositive(this float input,
		                                       float      factor) => Mathf.Round(f: input / factor) * factor;


		// ---------------------------------------------------------------------------------------------- Sign - Ref //

		/// <summary>
		/// 	Sets the float to a positive version of its value, regardless of its original signing.
		/// </summary>
		public static void ToPositive(ref this float input) => input = Mathf.Abs(f: input);


		/// <summary>
		/// 	Sets the float to a negative version of its value, regardless of its original signing.
		/// </summary>
		public static void ToNegative(ref this float input) => input = -Mathf.Abs(f: input);


		// ------------------------------------------------------------------------------------------ Sign - Returns //

		/// <summary>
		/// 	Returns a positive version of the input float.
		/// </summary>
		public static float GetPositive(this float input) => Mathf.Abs(f: input);


		/// <summary>
		/// 	Returns a negative version of the input float.
		/// </summary>
		public static float GetNegative(this float input) => -Mathf.Abs(f: input);

		// ------------------------------------------------------------------------------------------ Decimals - Ref //

		/// <summary>
		///     Returns the input with the given number of digits
		/// </summary>
		public static void ToDecimalPlaces(this float input,
		                                   int        digits) => input = (float) System.Math.Round(value: input,
		                                                                                           digits: digits);

		// -------------------------------------------------------------------------------------- Decimals - Returns //

		/// <summary>
		///     Returns the input with the given number of digits
		/// </summary>
		public static float GetWithDecimalPlaces(this float input,
		                                         int        digits) => (float) System.Math.Round(value: input,
		                                                                                         digits: digits);


		// ------------------------------------------------------------------------------------------ Rotation - Ref //
		
		public static void RadiansToDegrees(ref this float input) => input *= Mathf.Rad2Deg;

		public static void DegreesToRadians(ref this float input) => input *= Mathf.Deg2Rad;

		// -------------------------------------------------------------------------------------- Rotation - Returns //

		public static float GetRadiansToDegrees(this float input) => input * Mathf.Rad2Deg;

		public static float GetDegreesToRadians(this float input) => input * Mathf.Deg2Rad;

		// ----------------------------------------------------------------------------------- Differences - Returns //

		public static float GetDifference(this float input,
		                                          float      target) => Mathf.Abs(f: input - target);

		// ----------------------------------------------------------------------------------------- Noise - Returns //

		/// <summary>
		/// 	Returns a noise value given a 3D position
		/// </summary>
		public static float Perlin3D(float x,
		                             float y,
		                             float z) => ( Mathf.PerlinNoise(x: x,
		                                                             y: y) +
		                                           Mathf.PerlinNoise(x: y,
		                                                             y: z) +
		                                           Mathf.PerlinNoise(x: x,
		                                                             y: z) +
		                                           Mathf.PerlinNoise(x: y,
		                                                             y: x) +
		                                           Mathf.PerlinNoise(x: z,
		                                                             y: y) +
		                                           Mathf.PerlinNoise(x: z,
		                                                             y: x) ) / 6.0f;

	}

}