using System.Collections.Generic;
using System.Linq;

using UnityEngine;

namespace Vanilla.Core.Extensions
{

	public static class Vector2Extensions
	{

		// To Nearest


		public static void ToNearest(this Vector2 input,
		                             float        factor)
		{
			input.x.ToNearest(factor: factor);
			input.y.ToNearest(factor: factor);
		}


		public static Vector2 GetToNearest(this Vector2 input,
		                                   float        factor) => new Vector2(x: input.x.GetToNearest(factor: factor),
		                                                                       y: input.y.GetToNearest(factor: factor));


		public static void ToNearestPositiveOnly(this Vector2 input,
		                                         float        factor)
		{
			input.x.ToNearestPositive(factor: factor);
			input.y.ToNearestPositive(factor: factor);
		}


		public static Vector2 GetToNearestPositiveOnly(this Vector2 input,
		                                               float        factor)
		{
			input.x.GetNearestPositive(factor: factor);
			input.y.GetNearestPositive(factor: factor);

			return input;
		}


		//increase or decrease the length of vector by size
		public static Vector2 GetModifyMagnitude(this Vector2 input,
		                                         float        size)
		{
			//get the vector length
			var magnitude = input.magnitude;

			//calculate new vector length
			var newMagnitude = magnitude + size;

			//calculate the ratio of the new length to the old length
			var scale = newMagnitude / magnitude;

			//scale the vector
			return input * scale;
		}


		public static Vector2 GetChangeMagnitude(this Vector2 input,
		                                         float        newMagnitude) => input.normalized * newMagnitude;



		public static void ToRound(ref this Vector2 input)
		{
			input.x.ToRound();
			input.y.ToRound();
		}


		public static Vector2 GetRound(this Vector2 input) => new Vector2(x: Mathf.Round(f: input.x),
		                                                                  y: Mathf.Round(f: input.y));


		public static void ToCeiling(ref this Vector2 input)
		{
			input.x.ToCeiling();
			input.y.ToCeiling();
		}


		public static Vector2 GetCeiling(this Vector2 input) => new Vector2(x: Mathf.Ceil(f: input.x),
		                                                                    y: Mathf.Ceil(f: input.y));


		public static void ToFloor(ref this Vector2 input)
		{
			input.x.ToFloor();
			input.y.ToFloor();
		}


		public static Vector2 GetFloor(this Vector2 input) => new Vector2(x: Mathf.Floor(f: input.x),
		                                                                  y: Mathf.Floor(f: input.y));


		// --- Ranges, Wraps and Clamps --- //


		public static void RadialClamp(ref this Vector2 input,
		                               float            radius) => input = Vector2.ClampMagnitude(vector: input,
		                                                                                          maxLength: radius);


		public static Vector2 GetRadialClamp(this Vector2 input,
		                                     float        radius) => Vector2.ClampMagnitude(vector: input,
		                                                                                    maxLength: radius);


		public static void BoxClamp(ref this Vector2 input,
		                            float            size)
		{
			input.x.Clamp(min: -size,
			              max: size);

			input.y.Clamp(min: -size,
			              max: size);
		}


		public static Vector2 GetBoxClamp(this Vector2 input,
		                                  float        size) => new Vector3(x: input.x.GetClamp(min: -size,
		                                                                                        max: size),
		                                                                    y: input.y.GetClamp(min: -size,
		                                                                                        max: size));


		public static void RectangleClamp(ref this Vector2 input,
		                                  float            xMin,
		                                  float            xMax,
		                                  float            yMin,
		                                  float            yMax)
		{
			input.x.Clamp(min: xMin,
			              max: xMax);

			input.y.Clamp(min: yMin,
			              max: yMax);
		}


		public static Vector2 GetRectangleClamp(this Vector2 input,
		                                        float        xMin,
		                                        float        xMax,
		                                        float        yMin,
		                                        float        yMax) => new Vector2(x: input.x.GetClamp(min: xMin,
		                                                                                              max: xMax),
		                                                                          y: input.y.GetClamp(min: yMin,
		                                                                                              max: yMax));


		public static Vector2 GetRectangleClamp(this Vector2 input,
		                                        float        min,
		                                        float        max) => new Vector2(x: input.x.GetClamp(min: min,
		                                                                                             max: max),
		                                                                         y: input.y.GetClamp(min: min,
		                                                                                             max: max));


		/// <summary>
		///     Returns a Vector2 representing the sum of all the vectors in the collection.
		/// </summary>
		public static Vector2 Sum(this IEnumerable<Vector2> input) => input.Aggregate(seed: Vector2.zero,
		                                                                              func: (total,
		                                                                                     next) => total + next);


		/// <summary>
		///     Returns a Vector2 representing the average of all the vectors in the collection.
		/// </summary>
		public static Vector2 Average(this IEnumerable<Vector2> input) => input.Sum() / input.Count();


		// Rotate


		/// <summary>
		/// 	Rotates a given Vector2 by degrees.
		/// </summary>
		public static void Rotate(ref this Vector2 input,
		                          float            degrees)
		{
			var sin = Mathf.Sin(f: degrees * Mathf.Deg2Rad);
			var cos = Mathf.Cos(f: degrees * Mathf.Deg2Rad);

			var tx = input.x;
			var ty = input.y;

			input.x = cos * tx - sin * ty;
			input.y = sin * tx + cos * ty;
		}


		/// <summary>
		/// 	Rotate a Vector2
		/// </summary>
		///
		/// <param name="degrees">
		/// 	The amount of degrees to rotate the vector by
		/// </param>
		/// 
		/// <returns>
		/// 	A Vector2 representing the input vector rotated by degrees
		/// </returns>
		public static Vector2 GetRotate(this Vector2 input,
		                                float        degrees)
		{
			var sin = Mathf.Sin(f: degrees * Mathf.Deg2Rad);
			var cos = Mathf.Cos(f: degrees * Mathf.Deg2Rad);

			var tx = input.x;
			var ty = input.y;

			input.x = cos * tx - sin * ty;
			input.y = sin * tx + cos * ty;

			return input;
		}

	}

}